@@ -1,4 +1,4 @@
-# (c) Alexander Solovyov, 2009, under terms of the new BSD License
+# (c) Alexander Solovyov, 2009-2010, under terms of the new BSD License
 '''Command line arguments parser
 '''
 
@@ -6,12 +6,25 @@ import sys, traceback, getopt, types, textwrap, inspect, os
 from itertools import imap
 
 __all__ = ['command', 'dispatch']
-__version__ = '0.9.10'
+__version__ = '1.1'
 __author__ = 'Alexander Solovyov'
 __email__ = 'piranha@piranha.org.ua'
 
-write = sys.stdout.write
-err = sys.stderr.write
+try:
+    import locale
+    ENCODING = locale.getpreferredencoding()
+    if not ENCODING or ENCODING == 'mac-roman' or 'ascii' in ENCODING.lower():
+        ENCODING = 'UTF-8'
+except locale.Error:
+    ENCODING = 'UTF-8'
+
+def write(text, out=sys.stdout):
+    if isinstance(text, unicode):
+        return out.write(text.encode(ENCODING))
+    out.write(text)
+
+def err(text):
+    write(text, out=sys.stderr)
 
 CMDTABLE = {}
 
@@ -79,9 +92,10 @@ def command(options=None, usage=None, name=None, shortlist=False, hide=False):
             except Abort:
                 return -1
 
+            if opts.pop('help', False):
+                return help_func()
+
             try:
-                if opts.pop('help', False):
-                    return help_func()
                 return catcher(lambda: call_cmd(name_, func)(*args, **opts),
                                help_func)
             except Abort:
@@ -123,15 +137,23 @@ def dispatch(args=None, cmdtable=None, globaloptions=None,
     cmdtable['help'] = (help_(cmdtable, globaloptions), [], '[TOPIC]')
     help_func = cmdtable['help'][0]
 
-    autocomplete(cmdtable, args)
+    autocomplete(cmdtable, args, middleware)
 
     try:
         name, func, args, kwargs = catcher(
             lambda: _dispatch(args, cmdtable, globaloptions),
             help_func)
-        return catcher(
-            lambda: call_cmd(name, middleware(func))(*args, **kwargs),
-            help_func)
+    except Abort:
+        return -1
+
+    if name == '_completion':       # skip middleware
+        worker = lambda: call_cmd(name, func)(*args, **kwargs)
+    else:
+        worker = lambda: (call_cmd(name, middleware(func), depth=2)
+                          (*args, **kwargs))
+
+    try:
+        return catcher(worker, help_func)
     except Abort:
         return -1
 
@@ -155,9 +177,9 @@ def help_(cmdtable, globalopts):
 
             for cmd, info in cmdtable.items():
                 if cmd.startswith('~'):
-                    continue # do not display hidden commands
+                    continue  # do not display hidden commands
                 if shortlist and not cmd.startswith('^'):
-                    continue # short help contains only marked commands
+                    continue  # short help contains only marked commands
                 cmd = cmd.lstrip('^~')
                 doc = info[0].__doc__ or '(no help text available)'
                 hlp[cmd] = doc.splitlines()[0].rstrip()
@@ -169,7 +191,7 @@ def help_(cmdtable, globalopts):
             write('\ncommands:\n\n')
             for cmd in hlplist:
                 doc = hlp[cmd]
-                if False: # verbose?
+                if False:  # verbose?
                     write(' %s:\n     %s\n' % (cmd.replace('|', ', '), doc))
                 else:
                     write(' %-*s  %s\n' % (maxlen, cmd.split('|', 1)[0],
@@ -223,9 +245,7 @@ def help_cmd(func, usage, options):
     <BLANKLINE>
     '''
     write(usage + '\n')
-    doc = func.__doc__
-    if not doc:
-        doc = '(no help text available)'
+    doc = func.__doc__ or '(no help text available)'
     write('\n' + doc.strip() + '\n\n')
     if options:
         write(''.join(help_options(options)))
@@ -233,7 +253,8 @@ def help_cmd(func, usage, options):
 def help_options(options):
     yield 'options:\n\n'
     output = []
-    for short, name, default, desc in options:
+    for o in options:
+        short, name, default, desc = o[:4]
         if hasattr(default, '__call__'):
             default = default(None)
         default = default and ' (default: %s)' % default or ''
@@ -273,12 +294,14 @@ def parse(args, options):
     argmap, defmap, state = {}, {}, {}
     shortlist, namelist, funlist = '', [], []
 
-    for short, name, default, comment in options:
+    for o in options:
+        # might have the fifth completer element
+        short, name, default, comment = o[:4]
         if short and len(short) != 1:
-            raise FOError('Short option should be only a single'
-                          ' character: %s' % short)
+            raise OpsterError(
+                'Short option should be only a single character: %s' % short)
         if not name:
-            raise FOError(
+            raise OpsterError(
                 'Long name should be defined for every option')
         # change name to match Python styling
         pyname = name.replace('-', '_')
@@ -323,6 +346,8 @@ def parse(args, options):
         state[name] = defmap[name](None)
 
     return state, args
+
+
 # --------
 # Subcommand system
 # --------
@@ -412,8 +437,9 @@ def guess_options(func):
     args, varargs, varkw, defaults = inspect.getargspec(func)
     for name, option in zip(args[-len(defaults):], defaults):
         try:
-            sname, default, hlp = option
-            yield (sname, name.replace('_', '-'), default, hlp)
+            sname, default, hlp = option[:3]
+            completer = option[3] if len(option) > 3 else None
+            yield (sname, name.replace('_', '-'), default, hlp, completer)
         except TypeError:
             pass
 
@@ -438,33 +464,33 @@ def catcher(target, help_func):
         return target()
     except UnknownCommand, e:
         err("unknown command: '%s'\n" % e)
+        raise Abort()
     except AmbiguousCommand, e:
         err("command '%s' is ambiguous:\n    %s\n" %
             (e.args[0], ' '.join(e.args[1])))
+        raise Abort()
     except ParseError, e:
         err('%s: %s\n' % (e.args[0], e.args[1]))
         help_func(e.args[0])
+        raise Abort()
     except getopt.GetoptError, e:
-        err('error: %s\n' % e)
+        err('error: %s\n\n' % e)
         help_func()
-    except FOError, e:
+        raise Abort()
+    except OpsterError, e:
         err('%s\n' % e)
-    except KeyboardInterrupt:
-        err('interrupted!\n')
-    except SystemExit:
-        raise
-    except:
-        err('unknown exception encountered')
-        raise
+        raise Abort()
 
-    raise Abort
+def call_cmd(name, func, depth=1):
+    '''Wrapper for command call, catching situation with insufficient arguments
 
-def call_cmd(name, func):
+    ``depth`` is necessary when there is a middleware in setup
+    '''
     def inner(*args, **kwargs):
         try:
             return func(*args, **kwargs)
         except TypeError:
-            if len(traceback.extract_tb(sys.exc_info()[2])) == 1:
+            if len(traceback.extract_tb(sys.exc_info()[2])) == depth:
                 raise ParseError(name, "invalid arguments")
             raise
     return inner
@@ -491,7 +517,9 @@ def replace_name(usage, name):
 
 def sysname():
     name = sys.argv[0]
-    if name.startswith('./'):
+    if name.startswith('/'):
+        return name.rsplit('/', 1)[1]
+    elif name.startswith('./'):
         return name[2:]
     return name
 
@@ -513,20 +541,20 @@ except ImportError:
 # --------
 
 # Borrowed from PIP
-def autocomplete(cmdtable, args):
+def autocomplete(cmdtable, args, middleware):
     """Command and option completion.
 
     Enable by sourcing one of the completion shell scripts (bash or zsh).
     """
 
     # Don't complete if user hasn't sourced bash_completion file.
-    if not os.environ.has_key('OPSTER_AUTO_COMPLETE'):
+    if 'OPSTER_AUTO_COMPLETE' not in os.environ:
         return
     cwords = os.environ['COMP_WORDS'].split()[1:]
     cword = int(os.environ['COMP_CWORD'])
 
     try:
-        current = cwords[cword-1]
+        current = cwords[cword - 1]
     except IndexError:
         current = ''
 
@@ -540,18 +568,24 @@ def autocomplete(cmdtable, args):
 
     # command options
     elif cwords[0] in commands:
+        idx = -2 if current else -1
         options = []
         aliases, (cmd, opts, usage) = findcmd(cwords[0], cmdtable)
-        for (short, long, default, help) in opts:
-            options.append('-%s'  % short)
-            options.append('--%s' % long)
 
-        options = [o for o in options if o.startswith(current)]
-        print ' '.join(filter(lambda x: x.startswith(current), options))
+        for o in opts:
+            short, long, default, help = o[:4]
+            completer = o[4] if len(o) > 4 else None
+            short, long = '-%s' % short, '--%s' % long
+            options += [short, long]
+
+            if cwords[idx] in (short, long) and completer:
+                args = middleware(completer)(current)
+                print ' '.join(args),
+
+        print ' '.join((o for o in options if o.startswith(current)))
 
     sys.exit(1)
 
-
 COMPLETIONS = {
     'bash':
         """
@@ -582,31 +616,31 @@ compctl -K _opster_completion %s
     }
 
 @command(name='_completion', hide=True)
-def completion(type=('t', 'bash', 'Completion type (bash or zsh)')):
+def completion(type=('t', 'bash', 'Completion type (bash or zsh)'),
+               # kwargs will catch every global option, which we get
+               # anyway, because middleware is skipped
+               **kwargs):
     """Outputs completion script for bash or zsh."""
 
     prog_name = os.path.split(sys.argv[0])[1]
-    print COMPLETIONS[type] % prog_name
+    print COMPLETIONS[type].strip() % prog_name
 
 # --------
 # Exceptions
 # --------
 
 # Command exceptions
-class CommandException(Exception):
-    'Base class for command exceptions'
+class OpsterError(Exception):
+    'Base opster exception'
 
-class AmbiguousCommand(CommandException):
+class AmbiguousCommand(OpsterError):
     'Raised if command is ambiguous'
 
-class UnknownCommand(CommandException):
+class UnknownCommand(OpsterError):
     'Raised if command is unknown'
 
-class ParseError(CommandException):
+class ParseError(OpsterError):
     'Raised on error in command line parsing'
 
-class Abort(CommandException):
-    'Abort execution'
-
-class FOError(CommandException):
-    'Raised on trouble with opster configuration'
+class Abort(OpsterError):
+    'Processing error, abort execution'
